### skkdic.rb -- rskkserv module for skkdic format dictionary.

## Copyright (C) 1997-2000  Shugo Maeda
## Copyright (C) 2000,2001  YAMASHITA Junji

## Author:	Shugo Maeda <shugo@aianet.ne.jp>
## Maintainer:	YAMASHITA Junji <ysjj@unixuser.org>
## Version:	2.3.2b

## This file is part of rskkserv.

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 2, or (at
## your option) any later version.

## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.

### History:

## 2000/12/22:
# split from skkserv.rb.in

### Code:

class SKKDic
  include FileTest
  
  OKURI_ARI_LABEL = ";; okuri-ari entries.\n"
  OKURI_NASI_LABEL = ";; okuri-nasi entries.\n"
  
  OKURI_ARI_REGEXP_SJIS = /[\x81-\x9f\xe0-\xef][\x40-\x7e\x80-\xfc].*[a-z] $/n
  OKURI_ARI_REGEXP_EUCJP = /[\xa1-\xfe][\xa1-\xfe].*[a-z] $/n
  
  def initialize(path, cachedir, nocache, kcode)
    @path = path
    @file = open(@path)
    
    @cache_a = File.expand_path(File.basename(@path) + ".a", cachedir)
    @cache_n = File.expand_path(File.basename(@path) + ".n", cachedir)
    @nocache = nocache
    if kcode == :sjis
      @okuri_ari_regexp = OKURI_ARI_REGEXP_SJIS
    else
      @okuri_ari_regexp = OKURI_ARI_REGEXP_EUCJP
    end    

    if !@nocache &&
	(exist?(@cache_a) &&  File.ctime(@path) < File.ctime(@cache_a)) &&
	(exist?(@cache_n) &&  File.ctime(@path) < File.ctime(@cache_n))
      @okuri_ari_table = read_cache(@cache_a)
      @okuri_nasi_table = read_cache(@cache_n)
      if @okuri_ari_table.nil? || @okuri_nasi_table.nil?
	make_table
      end
    else
      make_table
    end
  end
  
  def make_table
    printf("Making index table for dictionary \`%s\'...\n", @path) if $OPT_d
      
    @okuri_ari_table = []
    @okuri_nasi_table = []
    for line in @file
      if line == OKURI_ARI_LABEL
	@okuri_ari_table.push(@file.pos)
	break
      end
    end
    for line in @file
      if line == OKURI_NASI_LABEL
	@okuri_nasi_table.push(@file.pos)
	break
      else
	@okuri_ari_table.push(@file.pos)
      end
    end
    @okuri_ari_table.pop
    @okuri_ari_table.reverse!
    for line in @file
      @okuri_nasi_table.push(@file.pos)
    end
    @okuri_nasi_table.pop
    
    unless @nocache
      write_cache(@okuri_ari_table, @cache_a)
      write_cache(@okuri_nasi_table, @cache_n)
    end
  end
  
  def write_cache(table, file)
    printf("Writing cache file %s ...\n", file) if $OPT_d
    open(file, "w") do |f|
      f.write(table.pack("i*"))
    end
  end
  
  def read_cache(file)
    printf("Reading cache file %s ...\n", file) if $OPT_d
    open(file) do |f|
      return f.read.unpack("i*")
    end
  end
  
  def search(kana_arg)
    printf("dic: %s, target: \"%s\", ", @path, kana_arg) if $OPT_d
    kana = kana_arg + " "
    
    if kana =~ @okuri_ari_regexp
      table = @okuri_ari_table
      print("okuri: ARI\n") if $OPT_d
    else
      table = @okuri_nasi_table
      print("okuri: NASI\n") if $OPT_d
    end
    
    left = 0
    right = table.length - 1
    while left <= right
      pos = (left + right) / 2

      @file.seek(table[pos], 0)
      buff = @file.readline
      print(buff) if $OPT_d
      case buff[0, kana.length] <=> kana
      when 0
	return buff[kana.length + 1 .. -3].split("/")
      when 1
	right = pos - 1
      when -1  
	left = pos + 1
      end
    end
    
    return []
  end

  def self.create_search_agents(serverconfig)
    config = serverconfig::SKKDic

    search_agents = [
      SKKDic.new(config::DICFILE, config::CACHEDIR, config::NOCACHE, config::KCODE)
    ]

    get_val = Proc.new {|m,key|
      m.const_defined?(key) ? m.const_get(key) : config.const_get(key)
    }
    # ruby 1.6.2, for get written order in file,
    # need reverse evaluation.
    config.constants.reverse_each do |ms|
      m = config.const_get(ms)
      next if m.type != Module

      search_agents.push(SKKDic.new(m.const_get("DICFILE"),
				    get_val.call(m, "CACHEDIR"),
				    get_val.call(m, "NOCACHE"),
				    get_val.call(m, "KCODE")))
    end

    search_agents
  end
end

# skkdic.rb ends here
